#include "xpcom-config.h"
#include "nsIGenericFactory.h"

#include "nspr.h" // debug

//#include "nsISupportsUtils.h"
#include "nsCOMPtr.h"
#include "nsXPCOM.h"
#include "nsIServiceManager.h"
#include "nsICategoryManager.h"

#include "nsMemory.h"
#include "nsIObserver.h"
#include "nsEmbedString.h"

// for NS_XPCOM_CURRENT_PROCESS_DIR ?
#include "nsDirectoryServiceDefs.h"

// more include include nsIContentPolicy (in unstable)
#include "nsIContentPolicy.h"
#include "nsIURI.h"
#include "nsIProperties.h"
#include "nsIFile.h"
#include "stdlib.h"
#include "nsISupportsPrimitives.h" // for nsISupportsCString
#include "nsIIOService.h" // for nsIIOService
#include "nsIComponentManager.h" // for nsIComponentManager

// include france interface
#include "nsIFrance.h"
// 894fe5e4-7467-42fb-8c6f-2a7515f99d63
#define FRANCE_CID \
{ 0x894fe5e4, 0x7467, 0x42fb, \
{ 0x8c, 0x6f, 0x2a, 0x75, 0x15, 0xf9, 0x9d, 0x63}}

#define FRANCE_CONTRACTID "@progysm.no-ip.org/xul/france"

// a simple link list
struct urlNode
{
  char* urlString;
  struct urlNode* next;
};


class France : public nsIObserver, public iFrance, public nsIContentPolicy
{
  public:
    France();
    virtual ~France();

    NS_DECL_ISUPPORTS
    NS_DECL_NSIOBSERVER
    NS_DECL_IFRANCE
    NS_DECL_NSICONTENTPOLICY

  private: // ???
    urlNode* mRootURLNode;
    PRBool mLocked;

    NS_IMETHOD readFile();
};

NS_IMPL_ISUPPORTS3(France, nsIObserver, iFrance, nsIContentPolicy);

France::France()
{
  //NS_INIT_ISUPPORTS()
  mLocked = PR_FALSE;
}

France::~France()
{
}


NS_IMETHODIMP
France::Observe(nsISupports *aSubject, 
                const char *aTopic,
                const PRUnichar *aData)
{
  readFile();
  return NS_OK;
}

NS_IMETHODIMP
France::readFile()
{
  nsCOMPtr<nsIServiceManager> servman;
  nsresult rv = NS_GetServiceManager(getter_AddRefs(servman));
  if (NS_FAILED(rv)) return rv;

  nsCOMPtr<nsIProperties> directoryService;
  rv = servman->GetServiceByContractID(NS_DIRECTORY_SERVICE_CONTRACTID,
                                       NS_GET_IID(nsIProperties),
                                       getter_AddRefs(directoryService));
  if (NS_FAILED(rv)) return rv;

  nsCOMPtr<nsIFile> appDir;
  rv = servman->GetServiceByContractID(NS_XPCOM_CURRENT_PROCESS_DIR,
                                       NS_GET_IID(nsIFile),
                                       getter_AddRefs(appDir));
  if (NS_FAILED(rv)) return rv;

  nsEmbedCString fileName("france.txt");
  appDir->AppendNative(fileName);
  return NS_OK;
}

NS_IMETHODIMP France::Lock()
{
  mLocked = PR_TRUE;
  return NS_OK;
}

NS_IMETHODIMP France::Unlock()
{
  mLocked = PR_FALSE;
  return NS_OK;
}

/* void addSite (in string url); */
NS_IMETHODIMP
France::AddSite(const char *url)
{
  urlNode* node = (urlNode*) malloc(sizeof(urlNode));
  node->urlString = strdup(url);
  node->next = mRootURLNode;
  mRootURLNode = node;
  return NS_OK;
}

NS_IMETHODIMP
France::RemoveSite(const char *url)
{
  urlNode* node = mRootURLNode;
  urlNode* prev = nsnull;

  while(node)
  {
    if (strcmp(node->urlString, url) == 0)
    {
      free(node->urlString);
      if (prev) 
        prev->next = node->next;
      free(node);
      return NS_OK;
    }
    prev = node;
    node = node->next;
  }
  return NS_ERROR_FAILURE;
}

NS_IMETHODIMP
France::SetSites(nsISimpleEnumerator * aSites)
{
  PRBool more = PR_TRUE;
  while(more) 
  {
    nsCOMPtr<nsISupports> supports;
    aSites->GetNext(getter_AddRefs(supports));

    nsCOMPtr<nsISupportsCString> supportsString = do_QueryInterface(supports);

    if (supportsString)
    {
      nsEmbedCString url;
      supportsString->GetData(url);
      AddSite(url.get());
    }

    aSites->HasMoreElements(&more);
  }
}

class myEnumerator : public nsISimpleEnumerator
{
  public:
    NS_DECL_ISUPPORTS
    NS_DECL_NSISIMPLEENUMERATOR

    myEnumerator(urlNode* node) {
      mNode = node;
    }
    virtual ~myEnumerator(void) {}

  protected:
    urlNode* mNode;
    nsCOMPtr<nsIComponentManager> mCompMgr;
};

NS_IMPL_ISUPPORTS1(myEnumerator, nsISimpleEnumerator);

NS_IMETHODIMP
myEnumerator::HasMoreElements(PRBool* aResult)
{
  if (!aResult)
    return NS_ERROR_NULL_POINTER;

  if (!mNode) {
    *aResult = PR_FALSE;
    return NS_OK;
  }
  *aResult = PR_TRUE;
  return NS_OK;
}

static NS_DEFINE_CID(kSupportsCStringCID, NS_SUPPORTS_CSTRING_CID);

NS_IMETHODIMP
myEnumerator::GetNext(nsISupports** aResult)
{
  if (!aResult)
    return NS_ERROR_NULL_POINTER;
  *aResult = nsnull;
  if (!mNode)
    return NS_ERROR_FAILURE;
  if (!mCompMgr)
  {
    NS_GetComponentManager(getter_AddRefs(mCompMgr));
    if(!mCompMgr)
      return NS_ERROR_UNEXPECTED;
  }

  nsISupportsCString* stringSupports;
  mCompMgr->CreateInstance(kSupportsCStringCID,
                           nsnull,
                           NS_GET_IID(nsISupportsCString),
                           (void**)&stringSupports);
  if(!stringSupports)
    return NS_ERROR_UNEXPECTED;

  nsEmbedCString str(mNode->urlString);
  stringSupports->SetData(str);

  *aResult = stringSupports; // addref'ed above
  mNode = mNode->next;
  return NS_OK;
}

NS_IMETHODIMP
France::GetSites(nsISimpleEnumerator * *aSites)
{
  myEnumerator* enumerator = new myEnumerator(mRootURLNode);
  if (!enumerator)
    return NS_ERROR_OUT_OF_MEMORY;
  NS_ADDREF(*aSites = enumerator);

  return NS_OK;
}

NS_IMETHODIMP
France::ShouldLoad(PRUint32 contentType,
                   nsIURI *contentLocation,
                   nsIURI *aRequestOrigin,
                   nsISupports *ctxt,
                   const nsACString & aMimeTypeGuess,
                   nsISupports *aExtra,
                   PRInt16 *_retval
                   )
{
  if (!contentLocation)
    return NS_ERROR_FAILURE;

  nsEmbedCString scheme;
  contentLocation->GetScheme(scheme);

  if (strcmp("http", scheme.get())  != 0 &&
      strcmp("https", scheme.get()) != 0 &&
      strcmp("ftp", scheme.get())   != 0)
  {
    *_retval = ACCEPT;
    return NS_OK;
  }

  nsEmbedCString hostToLoad;
  contentLocation->GetHost(hostToLoad);

  // assume failure until we declare it safe
  *_retval = REJECT_REQUEST;

  nsresult rv;

  urlNode* node = mRootURLNode;
  PRBool match = PR_FALSE;

  nsCOMPtr<nsIServiceManager> servMan;
  rv = NS_GetServiceManager(getter_AddRefs(servMan));
  if (NS_FAILED(rv)) return rv;

  nsCOMPtr<nsIIOService> mIOService;
  rv = servMan->GetServiceByContractID("@mozilla.org/network/io-service;1", 
                                       NS_GET_IID(nsIIOService),
                                       getter_AddRefs(mIOService));
  if (NS_FAILED(rv))
    return rv;

  while(node)
  {
    nsCOMPtr<nsIURI> uri;
    nsEmbedCString urlString(node->urlString);
    rv = mIOService->NewURI(urlString, nsnull, nsnull, getter_AddRefs(uri));

    if (NS_FAILED(rv))
      return rv;

    nsEmbedCString host;
    uri->GetHost(host);

    if (strcmp(hostToLoad.get(), host.get()) == 0)
    {
      *_retval = ACCEPT;
      return NS_OK;
    }
    node = node->next;
  }

  return NS_OK;
}

/* short shouldProcess (in unsigned long aContentType, in nsIURI aContentLocation, in nsIURI aRequestOrigin, in nsISupports aContext, in ACString aMimeType, in nsISupports aExtra); */
NS_IMETHODIMP France::ShouldProcess(PRUint32 aContentType, nsIURI *aContentLocation, nsIURI *aRequestOrigin, nsISupports *aContext, const nsACString & aMimeType, nsISupports *aExtra, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

static NS_METHOD FranceRegistration(nsIComponentManager *aCompMgr,
                                     nsIFile *aPath,
                                     const char *registryLocation,
                                     const char *componentType,
                                     const nsModuleComponentInfo *info)
{
  PR_fprintf(PR_STDOUT, "Chargement de France...\n");
  nsresult rv;

  nsCOMPtr<nsIServiceManager> servman = do_QueryInterface((nsISupports*)aCompMgr, &rv);
  if (NS_FAILED(rv))
    return rv;

  nsCOMPtr<nsICategoryManager> catman;
  rv = servman->GetServiceByContractID(NS_CATEGORYMANAGER_CONTRACTID,
                                       NS_GET_IID(nsICategoryManager),
                                       getter_AddRefs(catman));
  if (NS_FAILED(rv)) {
    return rv;
  }

  char *previous = nsnull;
  rv = catman->AddCategoryEntry("xpcom-startup",
                                "France",
                                FRANCE_CONTRACTID,
                                PR_TRUE,
                                PR_TRUE,
                                &previous);
  if (previous)
    nsMemory::Free(previous);

  rv = catman->AddCategoryEntry("content-policy",
                                "France",
                                FRANCE_CONTRACTID,
                                PR_TRUE,
                                PR_TRUE,
                                &previous);
  if (previous)
    nsMemory::Free(previous);

  return rv;
}

static NS_METHOD FranceUnregistration(nsIComponentManager *aCompMgr,
                                      nsIFile *aPath,
                                      const char *registryLocation,
                                      const nsModuleComponentInfo *info)
{
  nsresult rv;
  nsCOMPtr<nsIServiceManager> servman = do_QueryInterface((nsISupports*)aCompMgr, &rv);
  if (NS_FAILED(rv))
    return rv;

  nsCOMPtr<nsICategoryManager> catman;
  rv = servman->GetServiceByContractID(NS_CATEGORYMANAGER_CONTRACTID, 
                                       NS_GET_IID(nsICategoryManager),
                                       getter_AddRefs(catman));
  if (NS_FAILED(rv))
    return rv;

  rv = catman->DeleteCategoryEntry("xpcom-startup",
                                   "France",
                                   PR_TRUE);
  rv = catman->DeleteCategoryEntry("content-policy",
                                   "France",
                                   PR_TRUE);
  return rv;
}

NS_GENERIC_FACTORY_CONSTRUCTOR(France)

//NS_DECL_CLASSINFO(France)

static const nsModuleComponentInfo components[] = 
{
  {"France Class Name",
   FRANCE_CID,
   FRANCE_CONTRACTID,
   FranceConstructor,
   FranceRegistration,
   FranceUnregistration
  }
};


NS_IMPL_NSGETMODULE("FranceModule", components);

